#include "conf.h"

static PRStatus build_exporter_key(char* buf, const char* sec, const char* var, PRUint32 len)
{
    PRUint32 l;
    char* b;

    PL_strncpyz(buf, sec, len);

    l = PL_strnlen(buf, len);
    if (l > len - PL_strlen(var) - 2) /* account for ':' and null byte */
    {
        PR_LogPrint("section name length > %d chars: %s", len - PL_strlen(var) - 2, sec);
        return PR_FAILURE;
    }

    b = buf + l;
    *b++ = ':';
    PL_strcpy(b, var);

    return PR_SUCCESS;
}


static PRStatus do_exporter_section(void** wl, dictionary* d, const char* s)
{
    PRStatus status;
    PRUint32 len;
    char* var = "ipaddr";
    char* val;
    char key[32];

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("processing [%s] section", s));

    if (PR_FAILURE == build_exporter_key(key, s, var, sizeof(key)))
    {
        return PR_FAILURE;
    }

    val = iniparser_getstring(d, key, NULL);
    if (NULL == val)
    {
        PR_LogPrint("%s section has no %s variable", s, var);
        return PR_FAILURE;
    }
    PR_LogPrint("%s:%s = %s", s, var, val);

    status = whitelist_insert(wl, val);

    return status;
}


PR_IMPLEMENT(PRStatus) load_conf(conf_t* conf)
{
    int i, j;
    char *s;
    void* wl = NULL;
    void* owl;
    dictionary* confdict;
    PRNetAddrValue netAddrValue;

    /* init the conf file dictionary */
    confdict = iniparser_new(conf->fn);
    if (NULL == confdict)
    {
        PR_LogPrint("couldn't open conf file '%s'", conf->fn);
        return PR_FAILURE;
    }

#if 0
    if (PR_LOG_DEBUG <= globals.dlogger->level)
    {
        /* dump conf file dictonary to stderr */
        iniparser_dump(confdict, stdout);
    }
#endif

    PR_LogPrint("processing conf file '%s'", conf->fn);

    /* ensure the [global] section exists */
    if (!iniparser_find_entry(confdict, "global"))
    {
        PR_LogPrint("missing [global] section in %s", conf->fn);
        goto fail;
    }
    PR_LogPrint("processing global section in '%s'", conf->fn);

    /* process the [global] section in the conf file */

#ifdef DEBUG
    /* debug log level */
    globals.dlogger->level = iniparser_getint(confdict, "global:debugLoglevel", 0);
    if (globals.dlogger->level < 0 || globals.dlogger->level > 4)
    {
        PR_LogPrint("ignored bad global:debugLoglevel value %d in %s", globals.dlogger->level, conf->fn);
        globals.dlogger->level = 0;
    }
    PR_LogPrint("global:debugLoglevel = %d", globals.dlogger->level);
#endif

    /* number of exporters */
    conf->exporters = iniparser_getint(confdict, "global:exporters", 0);
    if (conf->exporters <= 0)
    {
        PR_LogPrint("bad global:exporters value (%d) in %s", conf->exporters, conf->fn);
        goto fail;
    }
    PR_LogPrint("global:exporters = %d", conf->exporters);

    /* number of threads */
    conf->threads = iniparser_getint(confdict, "global:threads", 0);
    if (conf->threads <= 0 || conf->threads > MAX_RECEIVE_THREADS)
    {
        PR_LogPrint("bad global:threads value (%d) in %s (must be > 0 and <= %d)",
                    conf->threads, conf->fn, MAX_RECEIVE_THREADS);
        goto fail;
    }
    /* ensure we have at least as many threads as exporters */
    if (conf->threads < conf->exporters)
    {
        conf->threads = 2 * conf->threads;
        PR_LogPrint("ignored globals:threads value in %s (must be >= globals:exporters)", conf->fn);
    }
    PR_LogPrint("globals:threads = %d", conf->threads);

    /* initialize the listen address and port */

    /* default is to listen on all interface addresses */
    PL_strcpy(conf->listen, "0.0.0.0");
    netAddrValue = PR_IpAddrAny;

    /* read listen address from the conf file [global] section */
    s = iniparser_getstring(confdict, "global:listen", NULL);
    if (NULL == s)
    {
        PR_LogPrint("missing global:listen value in %s", conf->fn);
        goto fail;
    }
    else
    {
        if (PL_strnlen(s) >= sizeof(conf->listen))
        {
            PR_LogPrint("ignored global:listen because its length exceeds %d chars",
                        sizeof(conf->listen));
        }
        else if (PR_SUCCESS == PR_StringToNetAddr(s, &conf->netAddr))
        {
            /* we set the address, so we don't want to muck with it when we set the port */
            PL_strcpy(conf->listen, s);
            netAddrValue = PR_IpAddrNull;
        }
        else
        {
            /* bad listen address specified in conf file */
            prerror("PR_StringToNetAddr");
        }
    }
    PR_LogPrint("global:listen = %s", conf->listen);

    /* read the listen port from the conf file [global] section */
    conf->port = iniparser_getint(confdict, "global:port", 0);
    if (conf->port <= 0)
    {
        /* we must have a port number */
        PR_LogPrint("bad or missing global:port value in %s", conf->fn);
        goto fail;
    }
    PR_LogPrint("global:port = %d", conf->port);

    /* complete the listen address/port initialization */
    if (PR_FAILURE == PR_InitializeNetAddr(netAddrValue, conf->port, &conf->netAddr))
    {
        prerror("PR_InitializeNetAddr");
        goto fail;
    }

    /* read the dbConnect string from the [global] section */
    s = iniparser_getstring(confdict, "global:dbConnect", NULL);
    if (NULL == s)
    {
        PR_LogPrint("missing global:dbConnect value in %s", conf->fn);
        goto fail;
    }
    if (PL_strnlen(s, sizeof(conf->dbConnect)) < sizeof(conf->dbConnect))
    {
        PL_strcpy(conf->dbConnect, s);
    }
    else
    {
        PR_LogPrint("ignored global:dbConnect (length exceeds %d chars)",
                    sizeof(conf->dbConnect));
    }
    PR_LogPrint("global:dbConnect = %s", conf->dbConnect);


    /* tcp session aggregation rate */
    conf->tcpSessAggHz = iniparser_getint(confdict, "global:tcpSessAggHz", 4);
    if (conf->tcpSessAggHz < 1 || conf->tcpSessAggHz > 10)
    {
        PR_LogPrint("bad global:tcpSessAggHz value (%d) in %s; must be 1 - 10",
                    conf->tcpSessAggHz, conf->fn);
        goto fail;
    }
    PR_LogPrint("global:tcpSessAggHz = %d", conf->tcpSessAggHz);

    /* udp session aggregation rate */
    conf->udpSessAggHz = iniparser_getint(confdict, "global:udpSessAggHz", 4);
    if (conf->udpSessAggHz < 1 || conf->udpSessAggHz > 10)
    {
        PR_LogPrint("bad global:udpSessAggHz value (%d) in %s; must be 1 - 10",
                    conf->udpSessAggHz, conf->fn);
        goto fail;
    }
    PR_LogPrint("global:udpSessAggHz = %d", conf->udpSessAggHz);

    /* other session aggregation rate */
    conf->otherSessAggHz = iniparser_getint(confdict, "global:otherSessAggHz", 4);
    if (conf->otherSessAggHz < 1 || conf->otherSessAggHz > 10)
    {
        PR_LogPrint("bad global:otherSessAggHz value (%d) in %s; must be 1 - 10",
                    conf->otherSessAggHz, conf->fn);
        goto fail;
    }
    PR_LogPrint("global:otherSessAggHz = %d", conf->otherSessAggHz);

    /* source traffic aggregation rate */
    conf->srcTraffAggHz = iniparser_getint(confdict, "global:srcTraffAggHz", 4);
    if (conf->srcTraffAggHz < 1 || conf->srcTraffAggHz > 10)
    {
        PR_LogPrint("bad global:srcTraffAggHz value (%d) in %s; must be 1 - 10",
                    conf->srcTraffAggHz, conf->fn);
        goto fail;
    }
    PR_LogPrint("global:srcTraffAggHz = %d", conf->srcTraffAggHz);

    /* destination traffic aggregation rate */
    conf->dstTraffAggHz = iniparser_getint(confdict, "global:dstTraffAggHz", 4);
    if (conf->dstTraffAggHz < 1 || conf->dstTraffAggHz > 10)
    {
        PR_LogPrint("bad global:dstTraffAggHz value (%d) in %s; must be 1 - 10",
                    conf->dstTraffAggHz, conf->fn);
        goto fail;
    }
    PR_LogPrint("global:dstTraffAggHz = %d", conf->dstTraffAggHz);

    /* whether to normalize flow times based on local system clock */
    i = iniparser_getboolean(confdict, "global:normalizeFlowTimes", 0);
    if (i)
    {
        conf->normalizeFlowTimes = PR_TRUE;
    }
    else
    {
        conf->normalizeFlowTimes = PR_FALSE;
    }
    PR_LogPrint("global:normalizeFlowTimes = %s",
                conf->normalizeFlowTimes ? "yes" : "no");

    /* process the individual exporter sections */

    /* ensure we have at least one exporter section */
    i = iniparser_getnsec(confdict);
    if (i < 2)
    {
        PR_LogPrint("missing exporter section(s) in %s", conf->fn);
        goto fail;
    }

    /* process all the exporter sections and and ensure we don't hit an unknown section name */
    j = 0;
    while (--i >= 0)
    {
        s = iniparser_getsecname(confdict, i);

        /* skip the already-processed global section */
        if (PL_strncmp(s, "global", 7) == 0)
        {
            continue;
        }

        /* we accept anything that begins with "exporter" as a valid section name */
        if (PL_strncmp(s, "exporter", 8) == 0)
        {
            if (PR_FAILURE == do_exporter_section(&wl, confdict, s))
            {
                goto fail;
            }
            j += 1;
            continue;
        }

        /* log an unknown section name and return */
        PR_LogPrint("unknown section name [%s] in %s", s, conf->fn);
        goto fail;
    }

    /* ensure number of exporter sections processed matches global:exporters */
    if (j != conf->exporters)
    {
        PR_LogPrint("global:exporters = %d in %s, but found %d exporter sections",
                    conf->exporters, conf->fn, j);
        goto fail;
    }

    /* activate the whitelist */
    owl = globals.conf.whitelist;
    globals.conf.whitelist = wl;

    /* destory the old whitelist (does nothing if NULL) */
    whitelist_destroy(owl);


#ifdef DEBUG
    if (PR_LOG_DEBUG <= globals.dlogger->level)
    {
        /* dump the whitelist values */
        whitelist_dump(&globals.conf.whitelist);
    }
#endif

    iniparser_free(confdict);

    return PR_SUCCESS;

fail:

    iniparser_free(confdict);

    return PR_FAILURE;
}
