#include "pfcapd.h"


PR_IMPLEMENT(void) nfv5_header_ntoh(host_header_t* dst, const nfv5_header_t* src)
{
    dst->version       = PR_ntohs(src->version);
    dst->count         = PR_ntohs(src->count);
    dst->sysUptime     = PR_ntohl(src->sysUptime);
    dst->unix_secs     = PR_ntohl(src->unix_secs);
    dst->unix_nsecs    = PR_ntohl(src->unix_nsecs);
    dst->flow_sequence = PR_ntohl(src->flow_sequence);
    dst->engine_type   = src->engine_type;
    dst->engine_ID     = src->engine_ID;
    dst->pad           = PR_ntohs(src->pad);
}


PR_IMPLEMENT(void) nfv5_record_ntoh(host_record_t* dst, const nfv5_record_t* src)
{
    /* don't alter ip addresses so they can be assigned to PRNetAddr's */
    /* dst->srcaddr   = PR_ntohl(src->srcaddr); */
    /* dst->dstaddr   = PR_ntohl(src->dstaddr); */
    /* dst->nexthop   = PR_ntohl(src->nexthop); */
    dst->input     = PR_ntohs(src->input);
    dst->output    = PR_ntohs(src->output);
    dst->dPkts     = PR_ntohl(src->dPkts);
    dst->dOctets   = PR_ntohl(src->dOctets);
    dst->first     = PR_ntohl(src->first);
    dst->last      = PR_ntohl(src->last);
    dst->srcport   = PR_ntohs(src->srcport);
    dst->dstport   = PR_ntohs(src->dstport);
    dst->pad1      = src->pad1;
    dst->tcp_flags = src->tcp_flags;
    dst->prot      = src->prot;
    dst->tos       = src->tos;
    dst->src_as    = PR_ntohs(src->src_as);
    dst->dst_as    = PR_ntohs(src->dst_as);
    dst->src_mask  = src->src_mask;
    dst->dst_mask  = src->dst_mask;
    dst->pad2      = PR_ntohs(src->pad2);
}


#ifndef DEBUG

#define EXPLODETIME(t,c,s)
#define LOG_HOST_HEADER(i,h)
#define LOG_HOST_RECORD(i,r)

#else

#define EXPLODETIME(t,c,s) PR_ExplodeTime(t,c,s);
#define LOG_HOST_HEADER(i,h) log_host_header(i,h)
#define LOG_HOST_RECORD(i,r) log_host_record(i,r)

static void log_host_header(const char* router, const host_header_t* hdr)
{
    char line[256];
    static const char* fmt =
        "router=%s"
        " version=%hu count=%hu"
        " sysUptime=%u unix_secs=%u unix_nsecs=%u"
        " flow_sequence=%u engine_type=%hhu engine_ID=%hhu";

    snprintf(line, sizeof(line), fmt,
             router,
             hdr->version, hdr->count,
             hdr->sysUptime, hdr->unix_secs, hdr->unix_nsecs,
             hdr->flow_sequence, hdr->engine_type, hdr->engine_ID);

    LOG_DEBUG(globals.dlogger, PR_LOG_TIMER1, (line));
}

void log_host_record(const char* router, const host_record_t* hrec)
{
    char srcaddr[16], dstaddr[16];
    char line[256];
    PRNetAddr netAddr;
    static const char* fmt =
        "router=%s"
        " src=%s:%hu dst=%s:%hu prot=%hhu tcp_flags=%hhu"
        " tos=%hhu first=%u last=%u in=%hu out=%hu"
        " dPkts=%u dOctets=%u"
        " src_as=%hu dst_as=%hu src_mask=%hhu dst_mask=%hhu";

    if (PR_FAILURE == PR_InitializeNetAddr(PR_IpAddrAny, 0, &netAddr))
    {
        prerror("netflow:process_nfv5_datagram:PR_InitializeNetAddr");
        return;
    }

    netAddr.inet.ip = hrec->srcaddr;
    if (PR_FAILURE == PR_NetAddrToString(&netAddr, srcaddr, sizeof(srcaddr)))
    {
        prerror("netflow:log_host_record:PR_NetAddrToString(srcaddr)");
        return;
    }

    netAddr.inet.ip = hrec->dstaddr;
    if (PR_FAILURE == PR_NetAddrToString(&netAddr, dstaddr, sizeof(dstaddr)))
    {
        prerror("netflow:log_host_record:PR_NetAddrToString(dstaddr)");
        return;
    }

    snprintf(line, sizeof(line), fmt,
             router,
             srcaddr, hrec->srcport, dstaddr, hrec->dstport, hrec->prot, hrec->tcp_flags,
             hrec->tos, hrec->first, hrec->last, hrec->input, hrec->output,
             hrec->dPkts, hrec->dOctets,
             hrec->src_as, hrec->dst_as, hrec->src_mask, hrec->dst_mask);
    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, (line));
}

#endif  /* DEBUG */


PR_IMPLEMENT(void) process_nfv5_datagram(db_data_t* dbData, const char* routerIP, nfv5_header_t* hdr, PRInt32 len)
{
    PRUint32 i;
    host_header_t hhdr;
    nfv5_record_t* rec;
    PRTime now = PR_Now();
    PRTime t, sentTime, bootTime;
    PRInt32 nfv5_len;
    PRUint32 nfv5_count;
#ifdef DEBUG
    PRExplodedTime explodedNowTime, explodedBootTime;
#endif

    nfv5_header_ntoh(&hhdr, hdr);
    LOG_HOST_HEADER(routerIP, &hhdr);

    if (5 != hhdr.version)
    {
        PR_LogPrint("v5 datagram rejected - version mismatch (%u)", hhdr.version);
        return;
    }

    nfv5_len = NFV5_HEADER_LEN + (NFV5_RECORD_LEN * hhdr.count);
    if (len != nfv5_len)
    {
        PR_LogPrint("datagram rejected - length mismatch between datagram (%u) and nfv5 data (%u)", len, nfv5_len);
        return;    /* we will compute flow times based on OUR time, not the router's -
     * NOT IMPLEMENTED, TODO: enable it based on a command line switch */

    }

    nfv5_count = hhdr.count;

    if (PR_TRUE == globals.conf.normalizeFlowTimes)
    {
        /* normalizing timestamps based on this system's clock, so just use
         * our time as the sent time */
        sentTime = now;
    }
    else
    {
        /* calculate sent time from datagram header unix_secs.unix_nses */
        LL_UI2L(sentTime, hhdr.unix_secs);           /* get sent time, sec */
        LL_MUL(sentTime, sentTime, PR_USEC_PER_SEC); /* convert sec to usec */
        LL_UI2L(t, hhdr.unix_nsecs);                 /* get residual nsec */
        LL_DIV(t, t, PR_NSEC_PER_USEC);              /* convert nsec to usec */
        LL_ADD(sentTime, sentTime, t);               /* compute sent time, usec */
    }
    /* calculate boot time as sent time - sysUptime */
    LL_UI2L(t, hhdr.sysUptime);                  /* get msec since boot time */
    LL_MUL(t, t, PR_USEC_PER_MSEC);              /* convert to usec */
    LL_SUB(bootTime, sentTime, t);               /* calculate boot time, usec */
#ifdef DEBUG
    PR_ExplodeTime(now, PR_GMTParameters, &explodedNowTime);
    PR_ExplodeTime(bootTime, PR_GMTParameters, &explodedBootTime);
    LOG_DEBUG(globals.dlogger, PR_LOG_TIMER1, ("router=%s now=%lld usec (%hd-%02d-%02d %02d:%02d:%02d.%06d), bootTime=%lld usec (%hd-%02d-%02d %02d:%02d:%02d.%06d)",
                                               routerIP, now,
                                               explodedNowTime.tm_year, explodedNowTime.tm_month+1, explodedNowTime.tm_mday,
                                               explodedNowTime.tm_hour, explodedNowTime.tm_min, explodedNowTime.tm_sec, explodedNowTime.tm_usec,
                                               bootTime,
                                               explodedBootTime.tm_year, explodedBootTime.tm_month+1, explodedBootTime.tm_mday,
                                               explodedBootTime.tm_hour, explodedBootTime.tm_min, explodedBootTime.tm_sec, explodedBootTime.tm_usec));
#endif
    rec = (void*)hdr + NFV5_HEADER_LEN;
    db_begin(dbData);
    for (i = 0; i < nfv5_count; i++)
    {
        nfv5_record_ntoh(rec, rec);
        LOG_HOST_RECORD(routerIP, rec);

        //db_begin(dbData);
        if (PR_FAILURE == db_insert(dbData, bootTime, routerIP, rec))
        {
            if  (i > 0)
            {
                db_rollback(dbData);
            }
            return;
        }
        //db_commit(dbData);
        rec += 1;
    }
    db_commit(dbData);
}
