#include <stdio.h>
#include <libgen.h>
#include <signal.h>
#include <plstr.h>
#include "pfcapd.h"

/* block all blockable, non-sync signals */
static PRStatus block_signals()
{
    sigset_t sigset;

    sigfillset(&sigset);        /* block everything except... */

    /* remove unblockable signals */
#ifdef SIGKILL
  sigdelset(&sigset, SIGKILL);
#endif
#ifdef SIGSTOP
  sigdelset(&sigset, SIGSTOP);
#endif
#ifdef SIGCONT
  sigdelset(&sigset, SIGCONT);
#endif
#ifdef SIGWAITING
  sigdelset(&sigset, SIGWAITING);
#endif

    /* remove sync signals */
#ifdef SIGABRT
  sigdelset(&sigset, SIGABRT);
#endif
#ifdef SIGBUS
  sigdelset(&sigset, SIGBUS);
#endif
#ifdef SIGEMT
  sigdelset(&sigset, SIGEMT);
#endif
#ifdef SIGFPE
  sigdelset(&sigset, SIGFPE);
#endif
#ifdef SIGILL
  sigdelset(&sigset, SIGILL);
#endif
#ifdef SIGIOT
  sigdelset(&sigset, SIGIOT);
#endif
#ifdef SIGPIPE
  sigdelset(&sigset, SIGPIPE);
#endif
#ifdef SIGSEGV
  sigdelset(&sigset, SIGSEGV);
#endif
#ifdef SIGSYS
  sigdelset(&sigset, SIGSYS);
#endif
#ifdef SIGTRAP
  sigdelset(&sigset, SIGTRAP);
#endif

    /* set the signal mask to block all possible signals */
    if (pthread_sigmask(SIG_BLOCK, &sigset, NULL) != 0)
    {
        perror("pthread_sigmask");
        return PR_FAILURE;
    }

    return PR_SUCCESS;
}


static void stop()
{
    PR_LogPrint("%s stop command received", globals.name);

    stop_threads(ts_stop);
}


static PRInt32 reloading = 0;

static void reload()
{
    PR_LogPrint("%s reload command received", globals.name);

    if (PR_FAILURE == load_conf(&globals.conf))
    {
        PR_LogPrint("%s terminating reload request due to conf file error",
                    globals.name);
        return;
    }

    stop_threads(ts_restart);

    PR_LogPrint("%s reload completed", globals.name);
}



/* Main-thread function to process control commands. For now, under Unix, it
 * just waits for a SIGTERM, sets globals.status = stop, then returns. Note
 * that this function assumes signals are already blocked because it's called
 * after the processing threads are created; signals must be blocked before
 * creating the threads so they inherit the signal mask. */
static void process_control_commands()
{
    int signo;
    sigset_t sigset;

    sigemptyset(&sigset);
    sigaddset(&sigset, SIGTERM);
    sigaddset(&sigset, SIGINT);
    while (1)
    {
        sigwait(&sigset, &signo);
        switch (signo)
        {
            case SIGTERM:
                stop();
                return;
            case SIGINT:
                reload();
                break;
            default:
                PR_LogPrint("aborting on receipt of unmasked signal %d", signo);
                abort();
        }
    }
}


static void usage()
{
    PR_fprintf(PR_STDERR,
               "\nUsage: %s [-h] | [-N] [-d] [-f conf_file]"
               "\n\n"
               "Options:"
               "\n\t-h            : print this help to stderr"
               "\n\t-d            : detach (become a daemon)"
               "\n\t-f conf_file  : conf file (default %s)"
               "\n\n",
               globals.name, globals.conf.fn);
}


static PRStatus init_udp()
{
    PRSocketOptionData socketOptions;

    /* allocate a new UDP socket */
    globals.udpSocket = PR_NewUDPSocket();
    if (NULL == globals.udpSocket)
    {
        prerror("init_udp:PR_NewUDPSocket");
        return PR_FAILURE;
    }

    /* bind the UDP socket to the IP and PORT specified in the conf file */
    if (PR_FAILURE == PR_Bind(globals.udpSocket, &globals.conf.netAddr))
    {
        prerror("init_udp:PR_Bind");
        return PR_FAILURE;
    }

    return PR_SUCCESS;
}


static PRIntn PR_CALLBACK pfcapd(PRIntn argc, char* argv[])
{
    int i;
    PRBool daemonize = PR_FALSE;
    PLOptStatus optStat;
    PLOptState* opt;
    PRThread* thread;
    PRExplodedTime startTime;
#ifdef DEBUG
    char* llend;
    long ll;
#endif

    /* set our name for log messages */
    globals.name = basename(argv[0]);

    /* set the default conf file name */
    PL_strncpyz(globals.conf.fn, DEFAULT_CONF, sizeof(globals.conf.fn));

    /* set up logging, controlled by the NSPR_LOG_MODULE environment variable */
    globals.logger = PR_NewLogModule("main");
    PR_ASSERT(globals.logger);

#ifdef DEBUG
    /* set up debug logging */
    globals.dlogger = PR_NewLogModule("debug");
    PR_ASSERT(globals.dlogger);
#endif

    /* process command line arguments */

    opt = PL_CreateOptState(argc, argv, "df:h");
    while (PL_OPT_EOL != (optStat = PL_GetNextOpt(opt)))
    {
        if (PL_OPT_BAD == optStat)
        {
            PR_fprintf(PR_STDERR, "\nBad option(s)\n");
            usage();
            return 1;
        }
        switch(opt->option)
        {
            case 'd':           /* daemonize */
                daemonize = PR_TRUE;
                break;

            case 'f':           /* config file */
                if (PL_strnlen(opt->value, sizeof(globals.conf.fn)) > MAX_CONF_STRLEN)
                {
                    PR_fprintf(PR_STDERR, "\nconf file name must be less than %d chars\n", MAX_CONF_STRLEN+1);
                    return 1;
                }
                PL_strcpy(globals.conf.fn, opt->value);
                break;

            case 'h':
                usage();
                return 0;
                break;

            default:            /* directory name to hold the anonymous file */
                PR_fprintf(PR_STDERR, "positional parameters not allowed: %s\n", opt->value);
                usage();
                return 1;
        }
    }
    PL_DestroyOptState(opt);

    /* log start time and options */
    PR_ExplodeTime(PR_Now(), PR_GMTParameters, &startTime);
    PR_LogPrint("%hd-%02d-%02d %02d:%02d:%02d.%06d GMT: started %s -f %s%s",
                startTime.tm_year, startTime.tm_month+1, startTime.tm_mday,
                startTime.tm_hour, startTime.tm_min, startTime.tm_sec, startTime.tm_usec,
                globals.name, globals.conf.fn, (daemonize ? " -d" : ""));

    LOG_DEBUG(globals.dlogger, PR_LOG_TIMER1, ("timer ops enabled at %u ticks/sec", PR_TicksPerSecond()));

    /* process the conf file */
    if (PR_FAILURE == load_conf(&globals.conf))
    {
        return 1;
    }

    if (daemonize)
    {
        DAEMONIZE(0,0);
    }

    if (PR_FAILURE == init_udp())
    {
        return 1;
    }

    /* create the data aggregation-related threads */
    if (PR_SUCCESS != start_threads())
    {
        PR_LogPrint("start_threads failed");
        return 1;
    }

    /* enter the control loop, which doesn't return until the shutdown command is received */
    process_control_commands();

    PR_LogPrint("%s exiting", globals.name);

    return 0;
}


int main(int argc, char* argv[])
{
    if (block_signals() == PR_FAILURE)
    {
        return 1;
    }

    /* spawn the primordial thread and wait for its return */
    PRIntn rc = PR_Initialize(pfcapd, argc, argv, 0);

    return rc;
}
