#include "config.h"
#include "threads.h"

/* the following sets the number of seconds an inactive thread sleeps before
 * waking up to check if it needs to act on a control command */
#define INACTIVE_THREAD_WAKEUP_SECS 1

/* age_*_session() update interval, in seconds */
#define AGE_SESSION_INTERVAL_SECS 60

/* aggregate thread references for the aggThreads vector array */
void aggregate_monitor_thread(void*);
void age_tcp_session_thread(void*);
void aggregate_tcp_session_thread(void*);
void age_udp_session_thread(void*);
void aggregate_udp_session_thread(void*);
void age_other_session_thread(void*);
void aggregate_other_session_thread(void*);
void aggregate_src_traffic_thread(void*);
void aggregate_dst_traffic_thread(void*);

static thread_vector_t aggThreads[] =
{
    { aggregate_monitor_thread,               "aggregate_monitor_thread",               NULL, ts_normal },
    { age_tcp_session_thread,                 "age_tcp_session_thread",                 NULL, ts_normal },
    { aggregate_tcp_session_thread,           "aggregate_tcp_session_thread",           NULL, ts_normal },
    { age_udp_session_thread,                 "age_udp_session_thread",                 NULL, ts_normal },
    { aggregate_udp_session_thread,           "aggregate_udp_session_thread",           NULL, ts_normal },
    { age_other_session_thread,               "age_other_session_thread",               NULL, ts_normal },
    { aggregate_other_session_thread,         "aggregate_other_session_thread",         NULL, ts_normal },
    { aggregate_src_traffic_thread,           "aggregate_src_traffic_thread",           NULL, ts_normal },
    { aggregate_dst_traffic_thread,           "aggregate_dst_traffic_thread",           NULL, ts_normal },
    /* vector with NULL values terminates the array */
    { NULL,                                   NULL,                                     NULL, ts_normal }
};

static thread_vector_t dgThreads[MAX_RECEIVE_THREADS];

/* current time, updated once per sec by aggregate_monitor_thread */
PR_IMPLEMENT(volatile PRTime) timeNow;

/* monitored aggregation variables */
static volatile PRInt64 topFlowId;   /* top flow id */
static volatile PRInt64 tcpFlowId;   /* aggregate tcp session flow id */
static volatile PRInt64 udpFlowId;   /* aggregate udp session flow id */
static volatile PRInt64 otherFlowId; /* aggregate other session flow id */
static volatile PRInt64 srcFlowId;   /* aggregate source traffic flow id */
static volatile PRInt64 dstFlowId;   /* aggregate destination traffic flow id */


/* generate a unique id for database connect; simply uses the
 * PR_GetCurrentThread() return value, assuming that each thread only uses a
 * single database connection
 *
 * Input buffer must be at least 21 bytes long (the max "printable" hex value
 * length of a 64-bit pointer 16 + len("CONN") + 1 null byte = 21)
 *
 * Returns id length on success, -1 on error
 */
static PRUint32 genConnectId(db_data_t* dbData)
{
    PRUint32 len;

    len = PR_snprintf(dbData->dbConnectId,
                      DBCONNECTIDMAXSZ,
                      "CONN%p", PR_GetCurrentThread());

    if (-1 == len)
    {
        PR_LogPrint("genConnectId: PR_snprintf failed");
        return -1;
    }

    return len;
}


void aggregate_monitor_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRIntervalTime timer, interval;
    PRIntervalTime minTicks = PR_TicksPerSecond(); /* this one kicks off once per second */
    PRExplodedTime xTimeNow;

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();

        db_begin(dbData);
        topFlowId = db_get_top_flow_id(dbData);
        db_commit(dbData);
        if (0LL == topFlowId)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }

        timeNow = PR_Now();
#ifdef DEBUG
        if (PR_LOG_TIMER2 <= globals.dlogger->level)
        {
            PR_ExplodeTime(timeNow, PR_GMTParameters, &xTimeNow);
            LOG_DEBUG(globals.dlogger, PR_LOG_TIMER2,
                      ("_monitor_ %hd-%02d-%02d %02d:%02d:%02d.%06d GMT:"
                       " fid=%lld tfid=%lld ufid=%lld ofid=%lld sfid=%lld dfid=%lld",
                       xTimeNow.tm_year, xTimeNow.tm_month+1, xTimeNow.tm_mday,
                       xTimeNow.tm_hour, xTimeNow.tm_min, xTimeNow.tm_sec, xTimeNow.tm_usec,
                       topFlowId, tcpFlowId, udpFlowId, otherFlowId, srcFlowId, dstFlowId));
        }
#endif
        if (threadVector->status != ts_normal)
        {

            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
        interval = PR_IntervalNow() - timer;
        if (interval < minTicks)
        {
            PR_Sleep(minTicks - interval);
        }
    }

    db_disconnect(dbData);
}


void age_tcp_session_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRStatus status;
    PRIntervalTime timer;
    PRIntervalTime wakeupTics = PR_TicksPerSecond() * INACTIVE_THREAD_WAKEUP_SECS;
    PRIntervalTime updateTics = PR_TicksPerSecond() * AGE_SESSION_INTERVAL_SECS;
#ifdef DEBUG
    PRIntervalTime _age_tcp_session_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_age_tcp_session_);
        db_begin(dbData);
        status = db_age_tcp_session(dbData);
        db_commit(dbData);
        STOP_TIMER2(_age_tcp_session_);
        if (PR_FAILURE == status)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        /* our update interval is long, so we do a loop to check every
         * INACTIVE_THREAD_WAKEUP_SECS whether we need to act on a control
         * command */
        do
        {
            if (threadVector->status != ts_normal)
            {
                PR_LogPrint("%s: returning due to %s request",
                            threadVector->name,
                            (threadVector->status == ts_stop) ? "stop" : "restart");
                goto done;
            }
            if (PR_IntervalNow() - timer >= updateTics)
            {
                break;
            }
            PR_Sleep(wakeupTics);
        } while(1);
    }

    done:
        db_disconnect(dbData);
}


void aggregate_tcp_session_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRIntervalTime timer, interval;
    PRIntervalTime minTicks = PR_TicksPerSecond() / globals.conf.tcpSessAggHz;
#ifdef DEBUG
    PRIntervalTime _aggregate_tcp_session_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_aggregate_tcp_session_);
        db_begin(dbData);
        tcpFlowId = db_aggregate_tcp_session(dbData);
        db_commit(dbData);
        STOP_TIMER2(_aggregate_tcp_session_);
        if (0LL == tcpFlowId)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        if (threadVector->status != ts_normal)
        {
            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
        interval = PR_IntervalNow() - timer;
        if (interval < minTicks)
        {
            PR_Sleep(minTicks - interval);
        }
    }

    db_disconnect(dbData);
}


void age_udp_session_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRStatus status;
    PRIntervalTime timer;
    PRIntervalTime wakeupTics = PR_TicksPerSecond() * INACTIVE_THREAD_WAKEUP_SECS;
    PRIntervalTime updateTics = PR_TicksPerSecond() * AGE_SESSION_INTERVAL_SECS;
#ifdef DEBUG
    PRIntervalTime _age_udp_session_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_age_udp_session_);
        db_begin(dbData);
        status = db_age_udp_session(dbData);
        db_commit(dbData);
        STOP_TIMER2(_age_udp_session_);
        if (PR_FAILURE == status)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        /* our update interval is long, so we do a loop to check every
         * INACTIVE_THREAD_WAKEUP_SECS whether we need to act on a control
         * command */
        do
        {
            if (threadVector->status != ts_normal)
            {
                PR_LogPrint("%s: returning due to %s request",
                            threadVector->name,
                            (threadVector->status == ts_stop) ? "stop" : "restart");
                goto done;
            }
            if (PR_IntervalNow() - timer >= updateTics)
            {
                break;
            }
            PR_Sleep(wakeupTics);
        } while(1);
    }

    done:
        db_disconnect(dbData);
}


void aggregate_udp_session_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRIntervalTime timer, interval;
    PRIntervalTime minTicks = PR_TicksPerSecond() / globals.conf.udpSessAggHz;
#ifdef DEBUG
    PRIntervalTime _aggregate_udp_session_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_aggregate_udp_session_);
        db_begin(dbData);
        udpFlowId = db_aggregate_udp_session(dbData);
        db_commit(dbData);
        STOP_TIMER2(_aggregate_udp_session_);
        if (0LL == udpFlowId)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        if (threadVector->status != ts_normal)
        {
            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
        interval = PR_IntervalNow() - timer;
        if (interval < minTicks)
        {
            PR_Sleep(minTicks - interval);
        }
    }

    db_disconnect(dbData);
}


void age_other_session_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRStatus status;
    PRIntervalTime timer;
    PRIntervalTime wakeupTics = PR_TicksPerSecond() * INACTIVE_THREAD_WAKEUP_SECS;
    PRIntervalTime updateTics = PR_TicksPerSecond() * AGE_SESSION_INTERVAL_SECS;
#ifdef DEBUG
    PRIntervalTime _age_other_session_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_age_other_session_);
        db_begin(dbData);
        status = db_age_other_session(dbData);
        db_commit(dbData);
        STOP_TIMER2(_age_other_session_);
        if (PR_FAILURE == status)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        /* our update interval is long, so we do a loop to check every
         * INACTIVE_THREAD_WAKEUP_SECS whether we need to act on a control
         * command */
        do
        {
            if (threadVector->status != ts_normal)
            {
                PR_LogPrint("%s: returning due to %s request",
                            threadVector->name,
                            (threadVector->status == ts_stop) ? "stop" : "restart");
                goto done;
            }
            if (PR_IntervalNow() - timer >= updateTics)
            {
                break;
            }
            PR_Sleep(wakeupTics);
        } while(1);
    }

    done:
        db_disconnect(dbData);
}


void aggregate_other_session_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRIntervalTime timer, interval;
    PRIntervalTime minTicks = PR_TicksPerSecond() / globals.conf.otherSessAggHz;
#ifdef DEBUG
    PRIntervalTime _aggregate_other_session_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_aggregate_other_session_);
        db_begin(dbData);
        otherFlowId = db_aggregate_other_session(dbData);
        db_commit(dbData);
        STOP_TIMER2(_aggregate_other_session_);
        if (0LL == otherFlowId)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        if (threadVector->status != ts_normal)
        {
            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
        interval = PR_IntervalNow() - timer;
        if (interval < minTicks)
        {
            PR_Sleep(minTicks - interval);
        }
    }

    db_disconnect(dbData);
}


void aggregate_src_traffic_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRIntervalTime timer, interval;
    PRIntervalTime minTicks = PR_TicksPerSecond() / globals.conf.srcTraffAggHz;
#ifdef DEBUG
    PRIntervalTime _aggregate_src_traffic_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_aggregate_src_traffic_);
        db_begin(dbData);
        srcFlowId = db_aggregate_src_traffic(dbData);
        db_commit(dbData);
        STOP_TIMER2(_aggregate_src_traffic_);
        if (0LL == srcFlowId)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        if (threadVector->status != ts_normal)
        {
            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
        interval = PR_IntervalNow() - timer;
        if (interval < minTicks)
        {
            PR_Sleep(minTicks - interval);
        }
    }

    db_disconnect(dbData);
}


void aggregate_dst_traffic_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRThread* thisThread = PR_GetCurrentThread();
    PRIntervalTime timer, interval;
    PRIntervalTime minTicks = PR_TicksPerSecond() / globals.conf.dstTraffAggHz;
#ifdef DEBUG
    PRIntervalTime _aggregate_dst_traffic_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    while(1)
    {
        timer = PR_IntervalNow();
        START_TIMER2(_aggregate_dst_traffic_);
        db_begin(dbData);
        dstFlowId = db_aggregate_dst_traffic(dbData);
        db_commit(dbData);
        STOP_TIMER2(_aggregate_dst_traffic_);
        if (0LL == dstFlowId)
        {
            PR_LogPrint("%s: returning due to database error", threadVector->name);
            break;
        }
        if (threadVector->status != ts_normal)
        {
            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
        interval = PR_IntervalNow() - timer;
        if (interval < minTicks)
        {
            PR_Sleep(minTicks - interval);
        }
    }

    db_disconnect(dbData);
}


void datagram_receive_thread(void* arg)
{
    thread_vector_t* threadVector = arg;
    db_data_t* dbData = &threadVector->dbData;
    PRInt32 bytes;
    PRNetAddr netAddr;
    PRIntervalTime recvExpireInterval = PR_TicksPerSecond() * INACTIVE_THREAD_WAKEUP_SECS;
    PRThread* thisThread = PR_GetCurrentThread();
    nfv5_header_t* dgbuf = PR_Malloc(UDP_BUFFER_SIZE);
    char addr[32];
    char routerIP[64];
    static char* globalIP = "0.0.0.0";
#ifdef DEBUG
    PRIntervalTime _datagram_;
#endif

    if (-1 == genConnectId(dbData))
    {
        PR_LogPrint("%s: returning due to genConnectId failure", threadVector->name);
        return;
    }

    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: started", threadVector->name));

    if (PR_FAILURE == db_connect(dbData))
    {
        PR_LogPrint("%s: returning due to db_connect failure", threadVector->name);
        return;
    }

    if (NULL == dgbuf)
    {
        prerror_abort("datagram_receive_thread: PR_Malloc");
    }

    while (1)
    {
        bytes = PR_RecvFrom(globals.udpSocket, dgbuf, UDP_BUFFER_SIZE, 0, &netAddr, recvExpireInterval);

        if (0 == bytes)
        {
            PR_LogPrint("%s: returning due to PR_RecvFrom error (network connection is closed)",
                        threadVector->name);
            break;
        }
        else if (-1 == bytes)
        {
            if (PR_GetError() == PR_IO_TIMEOUT_ERROR)
            {
                LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("%s: %d-second status check",
                                                          threadVector->name,
                                                          INACTIVE_THREAD_WAKEUP_SECS));
                goto statusCheck;
            }
            prerror("PR_RecvFrom");
            break;
        }

        START_TIMER1(_datagram_);

        if (NULL == whitelist_get(&globals.conf.whitelist, &netAddr))
        {

            if (PR_FAILURE == PR_NetAddrToString(&netAddr, addr, sizeof(addr)))
            {
                PL_strcpy(addr, "<unknown>");
            }
            PR_LogPrint("%s: rejecting packet from unknown %s", threadVector->name, addr);
            continue;
        }

        if (PR_FAILURE == PR_NetAddrToString(&netAddr, routerIP, sizeof(routerIP)))
        {
            PL_strcpy(routerIP, globalIP);
        }

        process_nfv5_datagram(dbData, routerIP, dgbuf, bytes);

        STOP_TIMER1_LOG_ROUTER(routerIP, _datagram_);

    statusCheck:
        if (threadVector->status != ts_normal)
        {
            PR_LogPrint("%s: returning due to %s request",
                        threadVector->name,
                        (threadVector->status == ts_stop) ? "stop" : "restart");
            break;
        }
    }

    db_disconnect(dbData);

    PR_Free(dgbuf);
}


static PRStatus start_thread(thread_vector_t* threadVector)
{
    if (NULL == threadVector->name)
    {
        PR_LogPrint("start_thread: thread name is uninitialized");
        return PR_FAILURE;
    }
    if (NULL == threadVector->func)
    {
        PR_LogPrint("start_thread: %s function is uninitialized", threadVector->name);
        return PR_FAILURE;
    }
    if (NULL != threadVector->thread)
    {
        PR_LogPrint("start_thread: %s is already started", threadVector->name);
        return PR_FAILURE;
    }
    threadVector->status = ts_normal;
    threadVector->thread = PR_CreateThread(PR_SYSTEM_THREAD,
                                           threadVector->func,
                                           threadVector,
                                           PR_PRIORITY_NORMAL,
                                           PR_GLOBAL_THREAD,
                                           PR_JOINABLE_THREAD,
                                           0);
    if (NULL == threadVector->thread)
    {
        prerror_abort("PR_CreateThread");
    }
    PR_LogPrint("start_thread: %s @0x%p started", threadVector->name, threadVector->thread);
    return PR_SUCCESS;
}


static PRStatus stop_thread(thread_vector_t* threadVector)
{
    if (NULL == threadVector->name)
    {
        PR_LogPrint("stop_thread: thread name is uninitialized");
        return PR_FAILURE;
    }
    if (NULL == threadVector->thread)
    {
        PR_LogPrint("stop_thread: %s is already stopped", threadVector->name);
        return PR_FAILURE;
    }
    PR_JoinThread(threadVector->thread);
    PR_LogPrint("stop_thread: %s @0x%p ended", threadVector->name, threadVector->thread);
    threadVector->thread = NULL;
    return PR_SUCCESS;
}


PR_IMPLEMENT(PRStatus) start_threads()
{
    int i;

    if (ts_normal != globals.runStatus)
    {
        PR_LogPrint("start_threads: run status indicates stop or restart in progress");
        return PR_FAILURE;
    }

    i = 0;
    while (aggThreads[i].func != NULL)
    {
        start_thread(&aggThreads[i++]);
    }

    for (i = 0; i < globals.conf.threads; i++)
    {
        dgThreads[i].func = datagram_receive_thread;
        dgThreads[i].name = "datagram_receive_thread";
        start_thread(&dgThreads[i]);
    }

    return PR_SUCCESS;
}


PR_IMPLEMENT(PRStatus) stop_threads(thread_status_t status)
{
    int i;

    if (ts_normal == status)
    {
        PR_LogPrint("stop_threads: status = normal is invalid (must be restart or stop)");
        return PR_FAILURE;
    }

    globals.runStatus = status;

    i = 0;
    while (aggThreads[i].func != NULL)
    {
        aggThreads[i].status = status;
        stop_thread(&aggThreads[i]);
        /* restart the aggregation thread if indicated */
        if (ts_restart == status)
        {
            start_thread(&aggThreads[i]);
        }
        ++i;
    }

    if (ts_stop == status)
    {
        /* for shutdown, notify all datagram receive threads at once to
         * return, so we don't have to serially wait for them to timeout if
         * the system is idle */
        for (i = 0; i < globals.conf.threads; i++)
        {
            dgThreads[i].status = status;
        }
        for (i = 0; i < globals.conf.threads; i++)
        {
            stop_thread(&dgThreads[i]);
        }
        return;
    }

    /* sanity check */
    if (ts_restart != status)
    {
        PR_LogPrint("stop_threads: aborting due to bad input thread status %d", status);
        abort();
    }

    /* for restart, serially notify each datagram receive thread to return,
     * then restart it immediately to minimize the impact of temporary thread
     * loss if the system is busy */
    for (i = 0; i < globals.conf.threads; i++)
    {
        dgThreads[i].status = status;
        /* the datagram receive thread count in the conf file might have been
         * increased, so check whether the end of the active threads has been
         * reached before reaching the thread count */
        if (NULL != dgThreads[i].thread)
        {
            stop_thread(&dgThreads[i]);
        }
        else
        {
            /* the receive count was increased if we reach this point, so
             * ensure the thread vector is properly initialized before
             * restarting it */
            dgThreads[i].func = datagram_receive_thread;
            dgThreads[i].name = "datagram_receive_thread";
        }
        start_thread(&dgThreads[i]);
    }

    /* if the datagram receive thread count in the conf file was reduced,
     * continue through the remaining thread vectors until we're sure all are
     * stopped */
    while (i < MAX_RECEIVE_THREADS)
    {
        if (NULL == dgThreads[i].thread)
        {
            /* first null thread value marks the end of active threads */
            break;
        }
        dgThreads[i].status = status;
        stop_thread(&dgThreads[i++]);
    }

    globals.runStatus = ts_normal;

    return PR_SUCCESS;
}
