#include <search.h>
#include "pfcapd.h"

#ifndef TDESTROY

/*$NetBSD: tdelete.c,v 1.2 1999/09/16 11:45:37 lukem Exp $*/

/*
 * Tree search generalized from Knuth (6.2.2) Algorithm T just like
 * the AT&T man page says.
 *
 * The node_t structure is for internal use only, lint doesn't grok it.
 *
 * Written by reading the System V Interface Definition, not the code.
 *
 * Totally public domain.
 */

/*
 * 2008.10.19 Modified by oban14@nothingworthwhile.org for the psyche project
 */

#include <stdlib.h>

typedef struct node {

    char         *key;

    struct node  *llink, *rlink;
    /*
     * The difference between this BSD and the GNU version of node_t is that
     * GNU defines an enum defined *after* the above fields, so the following
     * code should also work for old GNU libraries that may not have a
     * tdestroy() function.  -- oban14
     */

} node_t;

/* Walk the nodes of a tree */
static void trecurse(node_t *root, void (*free_action)(void *))
{
    if (root->llink != NULL)
        trecurse(root->llink, free_action);
    if (root->rlink != NULL)
        trecurse(root->rlink, free_action);

    (*free_action) ((void *) root->key);
    free(root);
}

static void tdestroy(void *vrootp, void (*freefct)(void *))
{
    node_t *root = (node_t *) vrootp;

    if (root != NULL)
        trecurse(root, freefct);
}

#endif  /* TDESTROY */


static int compare(const void* left, const void* right)
{
    if (((PRNetAddr*)left)->inet.ip == ((PRNetAddr*)right)->inet.ip)
    {
        return 0;
    }

    if (((PRNetAddr*)left)->inet.ip > ((PRNetAddr*)right)->inet.ip)
    {
        return 1;
    }

    return -1;
}


/* return pointer to whitelist netAddr that matches input value, NULL if not
 * found */
PR_IMPLEMENT(PRNetAddr*) whitelist_get(void** whitelist, PRNetAddr* netAddr)
{
    return (PRNetAddr*)tfind((void*)netAddr, whitelist, compare);
}


/* insert address into whitelist, log a message and return PR_FAILURE on error
 * (abort on memory allocation error), otherwise PR_SUCCESS */
PR_IMPLEMENT(PRStatus) whitelist_insert(void** whitelist, const char* addr)
{
    PRNetAddr* netAddr = PR_Malloc(sizeof(PRNetAddr));

    if (NULL == netAddr)
    {
        prerror_abort("whitelist:PR_Malloc()");
    }

    if (PR_FAILURE == PR_StringToNetAddr(addr, netAddr))
    {
        prerror("whitelist:PR_StringToNetAddr()");
        return PR_FAILURE;
    }

    if (NULL != whitelist_get(whitelist, netAddr))
    {
        PR_LogPrint("duplicate exporter ipaddr: %s", addr);
        return PR_FAILURE;
    }

    if (NULL == tsearch((void*)netAddr, whitelist, compare))
    {
        PR_LogPrint("whitelist_insert(%s) failed", addr);
        return PR_FAILURE;
    }

    return PR_SUCCESS;
}


static void dump(const void* node, const VISIT which, const int depth)
{
    PRNetAddr* netAddr;
    char addr[64];

    switch(which)
    {
        case preorder:
        case endorder:
            break;
        case postorder:
        case leaf:
            if (PR_SUCCESS == PR_NetAddrToString(*(PRNetAddr**)node, addr, sizeof(addr)))
            {
                LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, (addr));
            }
            else
            {
                prerror("whitelist:dump:PR_NetAddrToString");
            }
            break;
    }
}


PR_IMPLEMENT(void) whitelist_dump(void** whitelist)
{
    LOG_DEBUG(globals.dlogger, PR_LOG_DEBUG, ("dumping whitelist @%p", *whitelist));
    if (NULL == *whitelist)
    {
        return;
    }
    twalk(*whitelist, dump);
}


static void dummy_free_node(void* unused)
{
}


PR_IMPLEMENT(void) whitelist_destroy(void* whitelist_root)
{
    if (NULL != whitelist_root)
    {
        tdestroy(whitelist_root, dummy_free_node);
    }
}
